/** @file options.h
 *  @brief The header file for the command line option parser
 *  generated by GNU Gengetopt version 2.23
 *  http://www.gnu.org/software/gengetopt.
 *  DO NOT modify this file, since it can be overwritten
 *  @author GNU Gengetopt */

#ifndef OPTIONS_H
#define OPTIONS_H

/* If we use autoconf.  */
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdio.h> /* for FILE */

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#ifndef CMDLINE_PARSER_PACKAGE
/** @brief the program name (used for printing errors) */
#define CMDLINE_PARSER_PACKAGE "cssh"
#endif

#ifndef CMDLINE_PARSER_PACKAGE_NAME
/** @brief the complete program name (used for help and version) */
#define CMDLINE_PARSER_PACKAGE_NAME "cssh"
#endif

#ifndef CMDLINE_PARSER_VERSION
/** @brief the program version */
#define CMDLINE_PARSER_VERSION "0.0.2"
#endif

/** @brief Where the command line options are stored */
struct gengetopt_args_info
{
  const char *help_help; /**< @brief Print help and exit help description.  */
  const char *version_help; /**< @brief Print version and exit help description.  */
  const char *long_version_help; /**< @brief Full version and credentials help description.  */
  unsigned int verbose_min; /**< @brief Increase verbosity level (option can be given many times)'s minimum occurreces */
  unsigned int verbose_max; /**< @brief Increase verbosity level (option can be given many times)'s maximum occurreces */
  const char *verbose_help; /**< @brief Increase verbosity level (option can be given many times) help description.  */
  char * login_arg;	/**< @brief Login user (default: current user).  */
  char * login_orig;	/**< @brief Login user (default: current user) original value given at command line.  */
  const char *login_help; /**< @brief Login user (default: current user) help description.  */
  int port_arg;	/**< @brief Port to use for SSH (default: 22).  */
  char * port_orig;	/**< @brief Port to use for SSH (default: 22) original value given at command line.  */
  const char *port_help; /**< @brief Port to use for SSH (default: 22) help description.  */
  char * hosts_file_arg;	/**< @brief List of hosts to use in parallel (listed one per line in a file), optionally you can specify a port immediatelly after a colon after the hostname. Example 'user@host:port'.  */
  char * hosts_file_orig;	/**< @brief List of hosts to use in parallel (listed one per line in a file), optionally you can specify a port immediatelly after a colon after the hostname. Example 'user@host:port' original value given at command line.  */
  const char *hosts_file_help; /**< @brief List of hosts to use in parallel (listed one per line in a file), optionally you can specify a port immediatelly after a colon after the hostname. Example 'user@host:port' help description.  */
  const char *tagging_help; /**< @brief Enable tagging of each line with the name of the host help description.  */
  const char *recursive_help; /**< @brief Copy files and directories recursively help description.  */
  const char *progress_bar_help; /**< @brief Show progress bars during copy help description.  */
  int parallel_arg;	/**< @brief Number of files to transfer concurrently per host during a recursive copy (sftp protocol only; ignored for scp) (default='1').  */
  char * parallel_orig;	/**< @brief Number of files to transfer concurrently per host during a recursive copy (sftp protocol only; ignored for scp) original value given at command line.  */
  const char *parallel_help; /**< @brief Number of files to transfer concurrently per host during a recursive copy (sftp protocol only; ignored for scp) help description.  */
  char * protocol_arg;	/**< @brief Which SCP/SFTP transfer protocol to use for cscp (default='auto').  */
  char * protocol_orig;	/**< @brief Which SCP/SFTP transfer protocol to use for cscp original value given at command line.  */
  const char *protocol_help; /**< @brief Which SCP/SFTP transfer protocol to use for cscp help description.  */
  const char *ignore_connect_errors_help; /**< @brief Ignore host if we cannot connect to it at startup help description.  */
  const char *ignore_authentication_errors_help; /**< @brief Ignore host if we cannot authenticate to it at startup help description.  */
  const char *allow_password_authentication_help; /**< @brief Allow password identification if identification with keys fails help description.  */
  char * known_hosts_arg;	/**< @brief Use this known_hosts file instead of the default one.  */
  char * known_hosts_orig;	/**< @brief Use this known_hosts file instead of the default one original value given at command line.  */
  const char *known_hosts_help; /**< @brief Use this known_hosts file instead of the default one help description.  */
  
  unsigned int help_given ;	/**< @brief Whether help was given.  */
  unsigned int version_given ;	/**< @brief Whether version was given.  */
  unsigned int long_version_given ;	/**< @brief Whether long-version was given.  */
  unsigned int verbose_given ;	/**< @brief Whether verbose was given.  */
  unsigned int login_given ;	/**< @brief Whether login was given.  */
  unsigned int port_given ;	/**< @brief Whether port was given.  */
  unsigned int hosts_file_given ;	/**< @brief Whether hosts-file was given.  */
  unsigned int tagging_given ;	/**< @brief Whether tagging was given.  */
  unsigned int recursive_given ;	/**< @brief Whether recursive was given.  */
  unsigned int progress_bar_given ;	/**< @brief Whether progress-bar was given.  */
  unsigned int parallel_given ;	/**< @brief Whether parallel was given.  */
  unsigned int protocol_given ;	/**< @brief Whether protocol was given.  */
  unsigned int ignore_connect_errors_given ;	/**< @brief Whether ignore-connect-errors was given.  */
  unsigned int ignore_authentication_errors_given ;	/**< @brief Whether ignore-authentication-errors was given.  */
  unsigned int allow_password_authentication_given ;	/**< @brief Whether allow-password-authentication was given.  */
  unsigned int known_hosts_given ;	/**< @brief Whether known-hosts was given.  */

  char **inputs ; /**< @brief unnamed options (options without names) */
  unsigned inputs_num ; /**< @brief unnamed options number */
} ;

/** @brief The additional parameters to pass to parser functions */
struct cmdline_parser_params
{
  int override; /**< @brief whether to override possibly already present options (default 0) */
  int initialize; /**< @brief whether to initialize the option structure gengetopt_args_info (default 1) */
  int check_required; /**< @brief whether to check that all required options were provided (default 1) */
  int check_ambiguity; /**< @brief whether to check for options already specified in the option structure gengetopt_args_info (default 0) */
  int print_errors; /**< @brief whether getopt_long should print an error message for a bad option (default 1) */
} ;

/** @brief the purpose string of the program */
extern const char *gengetopt_args_info_purpose;
/** @brief the usage string of the program */
extern const char *gengetopt_args_info_usage;
/** @brief the description string of the program */
extern const char *gengetopt_args_info_description;
/** @brief all the lines making the help output */
extern const char *gengetopt_args_info_help[];

/**
 * The command line parser
 * @param argc the number of command line options
 * @param argv the command line options
 * @param args_info the structure where option information will be stored
 * @return 0 if everything went fine, NON 0 if an error took place
 */
int cmdline_parser (int argc, char **argv,
  struct gengetopt_args_info *args_info);

/**
 * The command line parser (version with additional parameters - deprecated)
 * @param argc the number of command line options
 * @param argv the command line options
 * @param args_info the structure where option information will be stored
 * @param override whether to override possibly already present options
 * @param initialize whether to initialize the option structure my_args_info
 * @param check_required whether to check that all required options were provided
 * @return 0 if everything went fine, NON 0 if an error took place
 * @deprecated use cmdline_parser_ext() instead
 */
int cmdline_parser2 (int argc, char **argv,
  struct gengetopt_args_info *args_info,
  int override, int initialize, int check_required);

/**
 * The command line parser (version with additional parameters)
 * @param argc the number of command line options
 * @param argv the command line options
 * @param args_info the structure where option information will be stored
 * @param params additional parameters for the parser
 * @return 0 if everything went fine, NON 0 if an error took place
 */
int cmdline_parser_ext (int argc, char **argv,
  struct gengetopt_args_info *args_info,
  struct cmdline_parser_params *params);

/**
 * Save the contents of the option struct into an already open FILE stream.
 * @param outfile the stream where to dump options
 * @param args_info the option struct to dump
 * @return 0 if everything went fine, NON 0 if an error took place
 */
int cmdline_parser_dump(FILE *outfile,
  struct gengetopt_args_info *args_info);

/**
 * Save the contents of the option struct into a (text) file.
 * This file can be read by the config file parser (if generated by gengetopt)
 * @param filename the file where to save
 * @param args_info the option struct to save
 * @return 0 if everything went fine, NON 0 if an error took place
 */
int cmdline_parser_file_save(const char *filename,
  struct gengetopt_args_info *args_info);

/**
 * Print the help
 */
void cmdline_parser_print_help(void);
/**
 * Print the version
 */
void cmdline_parser_print_version(void);

/**
 * Initializes all the fields a cmdline_parser_params structure 
 * to their default values
 * @param params the structure to initialize
 */
void cmdline_parser_params_init(struct cmdline_parser_params *params);

/**
 * Allocates dynamically a cmdline_parser_params structure and initializes
 * all its fields to their default values
 * @return the created and initialized cmdline_parser_params structure
 */
struct cmdline_parser_params *cmdline_parser_params_create(void);

/**
 * Initializes the passed gengetopt_args_info structure's fields
 * (also set default values for options that have a default)
 * @param args_info the structure to initialize
 */
void cmdline_parser_init (struct gengetopt_args_info *args_info);
/**
 * Deallocates the string fields of the gengetopt_args_info structure
 * (but does not deallocate the structure itself)
 * @param args_info the structure to deallocate
 */
void cmdline_parser_free (struct gengetopt_args_info *args_info);

/**
 * Checks that all the required options were specified
 * @param args_info the structure to check
 * @param prog_name the name of the program that will be used to print
 *   possible errors
 * @return
 */
int cmdline_parser_required (struct gengetopt_args_info *args_info,
  const char *prog_name);

extern const char *cmdline_parser_protocol_values[];  /**< @brief Possible values for protocol. */


#ifdef __cplusplus
}
#endif /* __cplusplus */
#endif /* OPTIONS_H */
